#include "BoidObjects.h"

#include "StringTable.h"
#include "RakPeerInterface.h"
#include "RakNetworkFactory.h"
#include <stdlib.h>
#include <stdio.h>
//#include "Kbhit.h"
#include <string.h>
#include "BitStream.h"
//#include "MessageIdentifiers.h"
//#include "NetworkIDManager.h"
//#include "RakSleep.h"
//#include "FormatString.h"
//#include "RakAssert.h"
//#include "StringCompressor.h"
#include "DS_List.h"

DataStructures::List<Boid*> myBoids;
DataStructures::List<Boid*> otherBoids;

// Boid parameters
float inertia_factor=0.1f;
float pull_factor=0.1f;
float proximity_distance=6.0f;
float proximity_factor=5.0f;
float bound_velocity=0.2f;
float max_velocity=5.0f;
Vec3f world_min_pt(-30.0f,0.0f,-30.0f);
Vec3f world_max_pt(30.0f,30.0f,30.0f);

extern bool isServer;
//extern NetworkIDManager networkIdManager;
//extern RakPeerInterface *rakPeer;
//extern ReplicaManager2 replicaManager;


Boid::Boid()
{
	systemAddress=UNASSIGNED_SYSTEM_ADDRESS;
	randomPos();
}

// Stop tracking this pointer
Boid::~Boid()
{
	unsigned int index = myBoids.GetIndexOf(this);
	if (index != MAX_UNSIGNED_LONG)
		myBoids.RemoveAtIndex(index);
	index = otherBoids.GetIndexOf(this);
	if (index != MAX_UNSIGNED_LONG)
		otherBoids.RemoveAtIndex(index);
}

void Boid::worldConstrain()
{
	if (pos[0] > world_max_pt[0]) 
		vel[0] += -bound_velocity;
	if (pos[0] < world_min_pt[0]) 
		vel[0] +=  bound_velocity;

	if (pos[1] > world_max_pt[1]) 
		vel[1] += -bound_velocity;
	if (pos[1] < world_min_pt[1]) 
		vel[1] += bound_velocity;

	if (pos[2] > world_max_pt[2]) 
		vel[2] += -bound_velocity;
	if (pos[2] < world_min_pt[2]) 
		vel[2] +=  bound_velocity;
};

void Boid::randomPos()
{
	pos[0] = world_min_pt[0]  + ((world_max_pt[0] - world_min_pt[0]) * (((float)rand())/RAND_MAX));
	pos[1] = world_min_pt[1]  + ((world_max_pt[1] - world_min_pt[1]) * (((float)rand())/RAND_MAX));
	pos[2] = world_min_pt[2]  + ((world_max_pt[2] - world_min_pt[2]) * (((float)rand())/RAND_MAX));
	vel = Vec3f(); // Set to null
};

bool Boid::SerializeConstruction(RakNet::BitStream *bitStream, SerializationContext *serializationContext)
{
	StringTable::Instance()->EncodeString("Boid",128,bitStream);
	return true;
}

bool Boid::Serialize(RakNet::BitStream *bitStream, SerializationContext *serializationContext)
{
	bitStream->Write(pos[0]);
	bitStream->Write(pos[1]);
	bitStream->Write(pos[2]);
	bitStream->Write(vel[0]);
	bitStream->Write(vel[1]);
	bitStream->Write(vel[2]);
	return true;
}

void Boid::Deserialize(RakNet::BitStream *bitStream, SerializationType serializationType, SystemAddress sender, RakNetTime timestamp)
{
	bitStream->Read(pos[0]);
	bitStream->Read(pos[1]);
	bitStream->Read(pos[2]);
	bitStream->Read(vel[0]);
	bitStream->Read(vel[1]);
	bitStream->Read(vel[2]);
	return;
}

bool Boid::QueryIsDestructionAuthority(void) const
{
	if (isServer) return true;
	Boid *boid = const_cast<Boid *>(this);
	return (myBoids.GetIndexOf(boid) != MAX_UNSIGNED_LONG);
}

bool Boid::QueryIsVisibilityAuthority(void) const
{
	return QueryIsDestructionAuthority();
}

bool Boid::QueryIsSerializationAuthority(void) const
{
	return QueryIsDestructionAuthority();
}

void Boid::DeleteBoidsByAddress( SystemAddress systemAddress )
{
	DataStructures::List<Boid*> toDelete;
	unsigned int i;

	// Two stages because destruction of a boid deletes it from otherBoid list, 
	// and deletion isn't iterator safe
	for (i=0; i < otherBoids.Size(); i++)
	{
		if (otherBoids[i]->systemAddress==systemAddress)
		{
			toDelete.Insert(otherBoids[i],  __FILE__, __LINE__);
		}
	}
	for (i=0; i < toDelete.Size(); i++)
	{
		toDelete[i]->BroadcastDestruction();
		delete toDelete[i];
	}
}

// See Connection_RM2::Construct in ReplicaManager2.h for a full explanation of each parameter
Replica2* ReplicaManager2BoidConnection::Construct(RakNet::BitStream *replicaData, SystemAddress sender, SerializationType type, ReplicaManager2 *replicaManager, RakNetTime timestamp, NetworkID networkId, bool networkIDCollision)
{
	char objectName[128];
	StringTable::Instance()->DecodeString(objectName,128,replicaData);
	if (networkIDCollision) return 0;
	if (strcmp(objectName,"Boid")==0)
	{
		Boid *newBoid = new Boid;
		newBoid->systemAddress = sender;
		otherBoids.Insert(newBoid,  __FILE__, __LINE__);	
		return newBoid;
	}
	return 0;
}




// Update a set of boids
void updateBoids(float time)
{
	unsigned int i,j;
	Vec3f center;
	Vec3f inertia;
	Vec3f pull;
	Vec3f diff;
	int proximate;

	for (i=0; i< myBoids.Size(); i++) 
	{
		center = myBoids[i]->pos + center; // Flock center
		inertia = myBoids[i]->vel + inertia; // Flock inertia
	}
	for (i=0; i< myBoids.Size(); i++) 
	{
		center = myBoids[i]->pos + center; // Flock center
		inertia = myBoids[i]->vel + inertia; // Flock inertia
	}

	center = center * (1.0f/((float)myBoids.Size() + otherBoids.Size())); // Take average to get flock center 
	inertia = inertia* (1.0f/((float)myBoids.Size() + otherBoids.Size())); // Take average to get flock inertia 
	inertia = inertia *inertia_factor * time; // Scale inertia, it is the same for each boid

	for (i=0; i< myBoids.Size(); i++) 
	{	
		Vec3f local_center;

		// Flock Cohesion 
		pull = center - myBoids[i]->pos;
		myBoids[i]->vel = (pull*pull_factor*time) + myBoids[i]->vel;

		// Flock Alignment 
		myBoids[i]->vel = (inertia) + myBoids[i]->vel; 

		// Neighbor Separation 
		proximate = 0; // Count the number of proximal boids
		for (j=0; j< myBoids.Size(); j++) 
		{
			if (i!=j) 
			{
				diff = myBoids[i]->pos - myBoids[j]->pos;
				if (diff.length() < proximity_distance) // These are the proximal ones
				{ 
					proximate++;
					local_center = diff + local_center;
				}
			}
		}
		for (j=0; j< otherBoids.Size(); j++) 
		{
			diff = myBoids[i]->pos - otherBoids[j]->pos;
			if (diff.length() < proximity_distance) // These are the proximal ones
			{ 
				proximate++;
				local_center = diff + local_center;
			}
		}

		if (proximate>0) // If there are proximal boids
			myBoids[i]->vel = (local_center*proximity_factor*time) + myBoids[i]->vel;

		// Max velocity 
		myBoids[i]->vel.limit(max_velocity);

		// Calculate new position
		myBoids[i]->pos = myBoids[i]->pos + (myBoids[i]->vel * time);

		// World limits 
		myBoids[i]->worldConstrain();
	}
}

void printAll()
{
	unsigned int i;
	if (!isServer)
	{
		printf("Number my boids %i:\n", myBoids.Size());
		for (i=0; i < myBoids.Size(); i++)
			printf("\t Boid %i. ptr=%p (%f,%f,%f)\n", i+1, myBoids[i],
			myBoids[i]->pos[0], myBoids[i]->pos[1], myBoids[i]->pos[2]);
	}

	printf("Number other boids %i:\n", otherBoids.Size());
	for (i=0; i < otherBoids.Size(); i++)
		printf("\t Boid %i. ptr=%p (%f,%f,%f)\n", i+1, otherBoids[i], 
		otherBoids[i]->pos[0], otherBoids[i]->pos[1], otherBoids[i]->pos[2]);

}
