#ifndef __BOID_OBJECTS_H
#define __BOID_OBJECTS_H

#include "ReplicaManager2.h"
#include "Vec3f.h"

using namespace RakNet;


class Boid : public Replica2
{
public:

	// System address of this user
	SystemAddress systemAddress;

	// Boid specific data
	Vec3f pos;
	Vec3f vel;

	// Boid constraints
	void worldConstrain();
	void randomPos();

 
	Boid();
	virtual ~Boid();

	virtual bool SerializeConstruction(RakNet::BitStream *bitStream, SerializationContext *serializationContext);
	virtual bool Serialize(RakNet::BitStream *bitStream, SerializationContext *serializationContext);
	virtual void Deserialize(RakNet::BitStream *bitStream, SerializationType serializationType, SystemAddress sender, RakNetTime timestamp);
	virtual bool QueryIsDestructionAuthority(void) const;
	virtual bool QueryIsVisibilityAuthority(void) const;
	virtual bool QueryIsSerializationAuthority(void) const;
	static void DeleteBoidsByAddress(SystemAddress systemAddress);
};


// One instance of Connection_RM2 is implicitly created per connection that uses ReplicaManager2. The most important function to implement is Construct() as this creates your game objects.
// It is designed this way so you can override per-connection behavior in your own game classes
class ReplicaManager2BoidConnection : public Connection_RM2
{
	// Callback used to create objects
	// See Connection_RM2::Construct in ReplicaManager2.h for a full explanation of each parameter
	Replica2* Construct(RakNet::BitStream *replicaData, SystemAddress sender, SerializationType type, ReplicaManager2 *replicaManager, RakNetTime timestamp, NetworkID networkId, bool networkIDCollision);
};

// This is a required class factory, that creates and destroys instances of ReplicaManager2BoidConnection
class ReplicaManager2BoidConnectionFactory : public Connection_RM2Factory {
	virtual Connection_RM2* AllocConnection(void) const {return new ReplicaManager2BoidConnection;}
	virtual void DeallocConnection(Connection_RM2* s) const {delete s;}
};

void updateBoids(float time);
void printAll();

#endif
