// Distributed Boids example in RakNet
// No visualisation
//
// Anthony Steed, 2009
//
// Originally based on the ReplicaManager2 example from RakNet
//

#include "StringTable.h"
#include "RakPeerInterface.h"
#include "RakNetworkFactory.h"
#include <stdio.h>
#include "Kbhit.h"
#include <string.h>
#include <ctype.h>
#include "BitStream.h"
#include "MessageIdentifiers.h"
#include "ReplicaManager2.h"
#include "NetworkIDManager.h"
#include "RakSleep.h"
#include "FormatString.h"
#include "StringCompressor.h"
#include "BoidObjects.h"

using namespace RakNet;

NetworkIDManager networkIdManager;
bool isServer;
RakPeerInterface *rakPeer;
ReplicaManager2 replicaManager;
ReplicaManager2BoidConnectionFactory connectionFactory;

// The arrays of boids
extern DataStructures::List<Boid*> myBoids;
extern DataStructures::List<Boid*> otherBoids;

int numberBoids=10;
int frequency=10;

int main(void)
{
	char ch;
	SocketDescriptor sd;
	char ip[128];
	static const int SERVER_PORT=12345;

	printf("Start as (c)lient or (s)erver? ");
	ch=_getche();

	rakPeer = RakNetworkFactory::GetRakPeerInterface();
	if (ch=='c' || ch=='C')
	{
		isServer=false;
		sd.port=0;
	}
	else
	{
		isServer=true;
		sd.port=SERVER_PORT;
	}

	rakPeer->SetNetworkIDManager(&networkIdManager);
	networkIdManager.SetIsNetworkIDAuthority(isServer);
	rakPeer->Startup(isServer ? 32 : 1,100,&sd,1);
	rakPeer->AttachPlugin(&replicaManager);
	replicaManager.SetAutoAddNewConnections(false);
	replicaManager.SetConnectionFactory(&connectionFactory);
	if (isServer)
		rakPeer->SetMaximumIncomingConnections(32);

	if (isServer==false)
	{
		printf("\nEnter server IP: ");
		gets(ip);
		if (ip[0]==0)
			strcpy(ip, "127.0.0.1");
		rakPeer->Connect(ip,SERVER_PORT,0,0,0);
		printf("Connecting...\n");
	}

	RakNet::StringTable::Instance()->AddString("Boid", false);

	if (isServer)
		printf("\nCommands:\n(Q)uit\n(P)rint out all objects\n");
	else
		printf("\nCommands:\n(Q)uit\n(P)rint out all objects\n(S)pawn / despawn Boid\n\n");

	// Enter infinite loop to run the system
	Packet *packet;
	bool quit=false;
	while (!quit)
	{
		updateBoids(1.0f/frequency);
		for (packet = rakPeer->Receive(); packet; rakPeer->DeallocatePacket(packet), packet = rakPeer->Receive())
		{
			switch (packet->data[0])
			{
			case ID_CONNECTION_ATTEMPT_FAILED:
				printf("ID_CONNECTION_ATTEMPT_FAILED\n");
				quit=true;
				break;
			case ID_NO_FREE_INCOMING_CONNECTIONS:
				printf("ID_NO_FREE_INCOMING_CONNECTIONS\n");
				quit=true;
				break;
			case ID_CONNECTION_REQUEST_ACCEPTED:
				printf("ID_CONNECTION_REQUEST_ACCEPTED\n");
				replicaManager.AddNewConnection(packet->systemAddress);
				break;
			case ID_NEW_INCOMING_CONNECTION:
				printf("ID_NEW_INCOMING_CONNECTION from %s\n", packet->systemAddress.ToString());
				replicaManager.AddNewConnection(packet->systemAddress);
				break;
			case ID_DISCONNECTION_NOTIFICATION:
				printf("ID_DISCONNECTION_NOTIFICATION\n");
				Boid::DeleteBoidsByAddress(packet->systemAddress);
				break;
			case ID_CONNECTION_LOST:
				printf("ID_CONNECTION_LOST\n");
				Boid::DeleteBoidsByAddress(packet->systemAddress);
				break;
			}
		}

		if (_kbhit())
		{
			ch = tolower(_getch());
			switch(ch)
			{
			case 'q':
				printf("Quitting.\n");
				quit=true;
				break;
			case 'p':
				printAll();
				break;
			case 's':
				if (!isServer)
				{
					if (myBoids.Size()!=0)
					{
						printf("Destroy Boids.\n");
						while (myBoids.Size()>0)
						{
							myBoids[0]->BroadcastDestruction();
							delete myBoids[0];
						}
					}
					else
					{
						printf("Create Boids.\n");
						for (int i=0;i<numberBoids;i++)
						{
							Boid* newBoid = new Boid();
							myBoids.Insert(newBoid,  __FILE__, __LINE__);
							newBoid->SetReplicaManager(&replicaManager);
							newBoid->AddAutoSerializeTimer(100);
							newBoid->BroadcastConstruction();
						}
					}
				}
				break;
			}
		}
		RakSleep((int)(1000.0/frequency));
	}

	rakPeer->Shutdown(100,0);
	RakNetworkFactory::DestroyRakPeerInterface(rakPeer);
}
